#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>

#include <semaphore.h>
#include <sys/mman.h>
#include <string.h>

#define checkError(expr,userMsg) \
    do {\
        if (!(expr)) {\
            perror(userMsg); \
            exit(EXIT_FAILURE);\
        }\
    } while (0)

#define BUF_SIZE (1024)

typedef struct {
    sem_t readSem;
    sem_t writeSem;
    char buffer[BUF_SIZE];
} Data_t;

void* getMemBlock(char* filePath, int* size) {

    int memFd = shm_open(filePath, O_RDWR | O_CREAT, 0600);
    checkError(memFd != -1, "shm_open open");

    struct stat fInfo;
    checkError(fstat(memFd, &fInfo) != -1, "ftruncate");
    *size = fInfo.st_size;

    void* addr = mmap(NULL, *size, PROT_READ|PROT_WRITE, MAP_SHARED, 
                        memFd, 0);
    checkError(addr != MAP_FAILED, "mmap");

    close(memFd);

    return addr;
}

int main(int argc, char** argv) {

    checkError(argc == 2, "args");

    int size = 0;
    Data_t* data = (Data_t*)getMemBlock(argv[1], &size);

    char buffer[BUF_SIZE];
    do {
        checkError(sem_wait(&(data->readSem)) != -1, "sem_wait");
        strcpy(buffer, data->buffer);
        checkError(sem_post(&(data->writeSem)) != -1, "sem-post");

        printf("%s\n", buffer);
        
    } while (strcmp(buffer, "quit") != 0);

    checkError(munmap(data, size) != -1, "munmap");

    exit(EXIT_SUCCESS);
}