#define _XOPEN_SOURCE (700)
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>

#include <semaphore.h>
#include <sys/mman.h>
#include <string.h>

#define checkError(expr,userMsg) \
    do {\
        if (!(expr)) {\
            perror(userMsg); \
            exit(EXIT_FAILURE);\
        }\
    } while (0)

#define BUF_SIZE (1024)

typedef struct {
    sem_t readSem;
    sem_t writeSem;
    char buffer[BUF_SIZE];
} Data_t;

void* createMemBlock(char* filePath, int size) {

    int memFd = shm_open(filePath, O_RDWR | O_CREAT, 0600);
    checkError(memFd != -1, "shm_open open");

    checkError(ftruncate(memFd, size) != -1, "ftruncate");

    void* addr = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, 
                        memFd, 0);
    checkError(addr != MAP_FAILED, "mmap");

    close(memFd);

    return addr;
}

int main(int argc, char** argv) {

    checkError(argc == 2, "args");

    Data_t* data = (Data_t*)createMemBlock(argv[1], sizeof(Data_t));

    checkError(sem_init(&(data->readSem), 1, 0) != -1, "sem_init");
    checkError(sem_init(&(data->writeSem), 1, 1) != -1, "sem_init");

    char buffer[BUF_SIZE];
    do {

        printf("Poruka za drugi proces: \n");
        scanf("%s", buffer);

        checkError(sem_wait(&(data->writeSem)) != -1, "sem_wait");
        strcpy(data->buffer, buffer);
        checkError(sem_post(&(data->readSem)) != -1, "sem_post");
        
    } while (strcmp(buffer, "quit") != 0);

    checkError(munmap(data, sizeof(Data_t)) != -1, "munmap");

    exit(EXIT_SUCCESS);
}